//  Code        :       Main, Brückenroboter
//  Version     :       1.0
//  Autoren     :       MM, LM
//  Funktion    :       Mainfunktion (ganzer Ablauf)


// ***Includes******************************************************************

#include "mbed.h"               // Stdlib
#include "EncoderCounter.h"     // ZAHW Encodercounter Lib


// ***Defines*******************************************************************

# define OK             1
# define notOK          0
# define PI             3.1416

#define ACCELERATION    3e-4        // Beschläunigung 
#define CpRMM           1560        // Encoderschritte Pro Umdrehung Main Motoren    
#define CpRFM           1560        // Encoderschritte Pro Umdrehung Front Motor   
#define WeelDM          50          // Durchmesser Haupträder in mm
#define WeelDF          50          // Durchmesser Fronträder in mm
#define WeeltoWeel      150         // Radabstand der Main Räder in mm

#define MotTuningL      1           // Motor links fine tuning (not greater then 1!)
#define MotTuningR      1           // Motor rechts fine tuning (not greater then 1!)
#define MotTuningF      1           // Motor vorne fine tuning (not greater then 1!)


// ***Initialisation************************************************************

DigitalOut myled(LED1);                     // Onboard Led (LED1) heisst "myled"
DigitalIn user_button(USER_BUTTON);         // Onboard Button (USER_BUTTON) heisst "user_button"

DigitalOut enable(PB_15);                   // If 1 then DCMotors Enabled

PwmOut pwm_motor1(PB_13);                   // PWM Pin MotorL
EncoderCounter counter1(PA_6, PC_7);        // Counter(pin A, pin B) MotorL

PwmOut pwm_motor2(PA_9);                    // PWM Pin MotorR
EncoderCounter counter2(PB_6, PB_7);        // Counter(pin A, pin B) MotorR

PwmOut pwm_motor3(PA_10);                   // PWM Pin MotorF
EncoderCounter counter3(PA_0, PA_1);        // Counter(pin A, pin B) MotorF

AnalogIn Sensor1(PC_2);
AnalogIn Sensor2(PC_3);
AnalogIn AnalogIn2(PC_5);
AnalogIn AnalogIn3(PB_1);

DigitalIn DigitalIn0(PC_9);
DigitalIn DigitalIn1(PC_8);
DigitalIn DigitalIn2(PC_6);
DigitalIn DigitalIn3(PB_12);


// ***Prototypen****************************************************************

int Systemcheck();

int PosCheck ();

int MoveStraight (int distance, int arm_motor_enable, int power);

int MoveTurn (int angle, int power, int mode);

int ArmsOut ();

int PartyMode();

int Motordrive (int power, int DistanceMotorL, int DistanceMotorR, int DistanceMotorF);

double CalculateVelocity (int MaxDistance, int Pos, double MaxVelocity);

int init();



// ***Globals*******************************************************************

int ProcessError = 0;



// ***Main**********************************************************************

int main ()
{
    printf("\nProgramm gestartet\r\n");
    init();                                     // Initialisiert alle Startwerte

    while (1)                                   // Main Loop
    {

        if (!user_button)                       // Startbutton gedrückt?
        {
            printf("\nStart\r\n");
            MoveStraight(1000,0,100);
            Motordrive(100, 1000, 0, 0);
            //MoveTurn(90,100, 1);
            //MoveTurn(90,100, 2);
            /*
            if (Systemcheck()== notOK)
            {
                ProcessError = 1;
                break;
            }
            
            MoveStraight(50,0,100);
            
            MoveTurn(90,100, 1);
            
            if (ArmsOut()==notOK)
            {
                ProcessError = 2;
                break;
            }
            
            MoveStraight(50,0,100);
            
            if (PosCheck()==notOK)              // PosCheck
            {
                ProcessError = 3;
                break;
            }   
            
            MoveStraight(200,1,20);
            MoveStraight(50,1,100);
            
            PartyMode();
            */
        }
    }
    
    if (ProcessError >= 1)
    {
        printf("***Error*** %d\r\n", ProcessError);
        while (1)
        {
            wait_ms(500);                       // Sleep 0.5s
            myled = 1;                          // LED is ON
            wait_ms(500);                       // Sleep 0.5s
            myled = 0;                          // LED is OFF
        }
    }
}



// ***Systemcheck***************************************************************

// Testet alle Bewegungen und Encoder

int Systemcheck()
{   
    int test_ok = 1; 

    //***Bewegungstest***

    if (MoveStraight(50, 0, 100)==notOK)
    {
           test_ok = 0; 
    }

    if (MoveStraight(-50, 0, 100)==notOK)
    {
           test_ok = 0; 
    }

    //***Arme testen***

    // ArmV drehen
    // ArmH drehen

    if (1) { // Taster nicht gedrückt
        while (0 || 0) { // Solange TasterV nicht gedrückt oder nicht länger als x Sekunden
            // ArmV drehen
        }
        while (0 || 0) { // Solange TasterH nicht gedrückt oder nicht länger als x Sekunden
            // ArmH drehen
        }
    }

    else {
        test_ok = 0;
    }

    if (0) { // Taster nicht gedrückt
        test_ok = 0;
    }

    //***Arm-Motor Testen***
    
    if (Motordrive(100, 0, 0, 100)==notOK)
    {
           test_ok = 0; 
    }
    
    if (test_ok > 0)    return OK;      // Test ok
    else                return notOK;   // Test not ok
    
}



// ***MoveStraight**************************************************************

// Fährt geradeaus
// Übergabewerte: Distanz, Armmotor Enablen, Geschwindigkeit (%)
    
int MoveStraight (int distance, int arm_motor_enable, int power)
{
    int DistanceMotorL = distance;
    int DistanceMotorR = distance;
    int DistanceMotorF = 0;
    
    if (arm_motor_enable == 1)
    {
        DistanceMotorF = distance;
    }
    
    if (Motordrive(power, DistanceMotorL, DistanceMotorR, DistanceMotorF)==OK)
    {
        return OK;
    }
    else return notOK;
}



// ***MoveTurn******************************************************************

// Fährt eine Kurve
// Übergabewerte: Winkel, Geschwindigkeit (%)

int MoveTurn (int angle, int power, int mode)
{
    int distance1 = (WeeltoWeel*PI*abs(angle))/360; // distanz = 2*r*pi*angle/360
    int distance2 = 0;
    int DistanceMotorL;
    int DistanceMotorR;

    if (mode ==2)
    {
        distance1=distance1/2;
        distance2=-distance1;
    }

    if (0>angle)
    {
        DistanceMotorL = distance1;
        DistanceMotorR = distance2;
    }
    else
    {
        DistanceMotorL = distance2;
        DistanceMotorR = distance1;
    }
    
    if (Motordrive(power, DistanceMotorL, DistanceMotorR, 0)==OK)
    {
        return OK;
    }
    else return notOK;
}



// ***Motordrive************************************************************************

// Kontrolliert die Motoren
// Übergabewerte: Max Power + welche Motoren davon betroffen sind

int Motordrive (int power, int DistanceMotorL, int DistanceMotorR, int DistanceMotorF)
{     
    double max_velocity = (double)power/200;
    double velocityL, velocityR, velocityF;
    double StarterL, StarterR, StarterF;
    int PosL =0;
    int PosR =0;
    int PosF =0;
    int PosChangeL, PosChangeR, PosChangeF;
    int dirL, dirR, dirF;
    int EncOldL, EncOldR, EncOldF;
    int EncNowL, EncNowR, EncNowF;
    int NotDone = 1;
    
    if (DistanceMotorL<0)   { dirL=-1;  }   else { dirL=1; }
    if (DistanceMotorR<0)   { dirR=-1;  }   else { dirR=1; }
    if (DistanceMotorF<0)   { dirF=-1;  }   else { dirF=1; }
    
    DistanceMotorL = (int)((abs(DistanceMotorL))/(WeelDM*PI) * CpRMM);
    DistanceMotorR = (int)((abs(DistanceMotorR))/(WeelDM*PI) * CpRMM);
    DistanceMotorF = (int)((abs(DistanceMotorF))/(WeelDF*PI) * CpRFM);
    
    EncNowL = counter1.read();
    EncNowR = counter2.read();
    EncNowF = counter3.read();
    EncOldL = EncNowL;
    EncOldR = EncNowR;
    EncOldF = EncNowF;
    StarterL=0.05;
    StarterR=0.05;
    StarterF=0.05;
    
    
    enable = 1;
    
    printf("DistanceMotorR = %d, DistanceMotorR = %d\r\n", DistanceMotorL, DistanceMotorR);
    
    while (NotDone)
    {   
        NotDone=0;
        
        EncNowL = counter1.read();
        EncNowR = counter2.read();
        EncNowF = counter3.read();
        
        PosChangeL = abs(abs(EncNowL)-abs(EncOldL));
        PosChangeR = abs(abs(EncNowR)-abs(EncOldR));
        PosChangeF = abs(abs(EncNowF)-abs(EncOldF));
        
        PosL = PosL + PosChangeL;
        PosR = PosR + PosChangeR;
        PosF = PosF + PosChangeF;
        
        EncOldL = EncNowL;
        EncOldR = EncNowR;
        EncOldF = EncNowF;
        
        if (DistanceMotorL > PosL)
        {
            NotDone++;
            if (PosChangeL <= 30)   StarterL=StarterL+0.002;  
            else { if (StarterL>0)  StarterL=StarterL-0.002;    }
            velocityL = MotTuningL * CalculateVelocity(DistanceMotorL, PosL, max_velocity) + StarterL;
            pwm_motor1.write(0.5 + velocityL * dirL);
        }
        else { pwm_motor1.write(0.5); }


        if (DistanceMotorR > PosR)
        {
            NotDone++;
            if (PosChangeR <= 30)   StarterR=StarterR+0.002;
            else { if (StarterR>0)  StarterR=StarterR-0.002;    } 
            velocityR = MotTuningR * CalculateVelocity(DistanceMotorR, PosR, max_velocity) + StarterR;
            pwm_motor2.write(0.5 + velocityR * dirR);
        }
        else { pwm_motor2.write(0.5); }
        
        
        if (DistanceMotorF > PosF)
        {
            NotDone++;
            if (PosChangeF <= 30)   StarterF=StarterF+0.002;
            else { if (StarterF>0)  StarterF=StarterF-0.002;    }
            velocityF = MotTuningF * CalculateVelocity(DistanceMotorF, PosF, max_velocity) + StarterF;
            pwm_motor3.write(0.5 + velocityF * dirF);
        }
        else { pwm_motor3.write(0.5); }
        
        
        printf("velocityL = %lf, velocityR = %lf pos_aktuellL = %d, pos_aktuellR = %d\r\n", velocityL, velocityR, PosL, PosR);
        
        wait(0.02f);    // Takt 0.02s, 50Hz
    }
    
    enable = 0;
    
    if (abs(DistanceMotorL-PosL)<20 && abs(DistanceMotorR-PosR)<20 && abs(DistanceMotorF-PosF)<20)
    {
        return OK;
    }
    else return notOK;
}



// ***PosCheck******************************************************************

// Kontrolliert, ob die Roboter Parallel zur Schlucht ist

int PosCheck()
{
    int PosOk = 0;
    int State  = 0;
    int StateOld  = 0;
    enable = 1;
    
    while (PosOk <= 5)
    {   
        State = 0;
        if (Sensor1.read()>=(double)0.5) State = 10;
        if (Sensor2.read()>=(double)0.5) State++;
        if (State != StateOld) PosOk++;
        StateOld=State;
        
        switch (State)
        {
            case 00:       // Kein Sensor erkennt die Schlucht
            
                pwm_motor1.period(0.0006f);
                pwm_motor2.period(0.0006f);
                wait(0.02f);
                
            break;
            
            case 11:       // Beide Sensoren erkennen die Schlucht
            
                pwm_motor1.period(0.0004f);
                pwm_motor2.period(0.0004f);
                wait(0.02f);
                
            break;
            
            case 1:        // Linker Sensor erkennt die Schlucht
            
                pwm_motor1.period(0.0005f);
                pwm_motor2.period(0.0004f);
                wait(0.02f);
                
            break;
            
            case 10:       // Rechter Sensor erkennt die Schlucht
            
                pwm_motor1.period(0.0004f);
                pwm_motor2.period(0.0005f);
                wait(0.02f);
                
            break;
        }
    }
    enable = 0;
    return 1;
}



// ***ArmsOut*******************************************************************

// Fährt den Brückenmechanismus aus

int ArmsOut()
{
    return 1; 
}



// ***PartyMode*****************************************************************

// Feiert den Sieg

int PartyMode()
{
    return 1; 
}



// ***CalculateVelocity*********************************************************

// Berechnet die Momentane Sollgeschwindigkeit
// Übergabewerte: Max Power + welche Motoren davon betroffen sind

double CalculateVelocity(int MaxDistance, int Pos, double MaxVelocity)
{
    double velocity = 0;
    
    if ((Pos*2) <= MaxDistance)
    {
        if (Pos*ACCELERATION >= MaxVelocity)
        {
            velocity = MaxVelocity;
        }
        
        else
        {
            velocity = Pos*ACCELERATION;
        }
    }
    else
    {
        if ((MaxDistance-Pos)*ACCELERATION >= MaxVelocity)
        {
            velocity = MaxVelocity;
        }
        else
        {
            velocity = (MaxDistance-Pos)*ACCELERATION;
        }
    }
    return velocity;
}


 
// ***Init**********************************************************************
// Initialisiert alle Startwerte

int init()
{
    // initialise DC Motor PWM
    pwm_motor1.period(0.0005f); // 0.5ms 2KHz
    pwm_motor1.write(0.5f);     // %DutyCycle -> keine Bewegung
    pwm_motor2.period(0.0005f); // 0.5ms 2KHz
    pwm_motor2.write(0.5f);     // %DutyCycle -> keine Bewegung
    pwm_motor3.period(0.0005f); // 0.5ms 2KHz
    pwm_motor3.write(0.5f);     // %DutyCycle -> keine Bewegung
    
    printf("\nMotoren Initialisiert\r\n");
    
    return 1;
}